#!/usr/bin/env python
#
# togglepanel.py - A panel which contains a button, and some content.
# Pushing the button toggles the visibility of the content.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TogglePanel` class, which is a panel that
contains a button and some content. Pushing the button toggles the visibility
of the content.
"""


import warnings
import six

import wx
import wx.lib.newevent as wxevent


class TogglePanel(wx.Panel):
    """A :class:`TogglePanel` is a ``wx.Panel`` that contains a button and
    some content.

    Pushing the button toggles the visibility of the content.


    All of the content should be added to the panel which is returned by the
    :meth:`GetPane` method.


    The ``TogglePanel`` may be used in place of the ``wx.CollapsiblePane``,
    which is buggy under Linux/GTK.
    """


    def __init__(self,
                 parent,
                 toggleSide=wx.TOP,
                 initialState=True,
                 label=None):
        """Create a :class:`TogglePanel`.

        :arg parent:       The :mod:`wx` parent object.

        :arg toggleSide:   Which side to place the toggle button. Must be one
                           of :attr:`wx.TOP`, :attr:`wx.BOTTOM`,
                           :attr:`wx.LEFT`, or :attr:`wx.RIGHT`.

        :arg initialState: Initial state for the panel content - visible
                           (``True``) or hidden (``False``).

        :arg label:        A label to be displayed on the toggle button.
        """

        wx.Panel.__init__(self, parent)

        self.__contentPanel = wx.Panel(self)

        if toggleSide in (wx.TOP, wx.BOTTOM):
            self.__mainSizer   = wx.BoxSizer(wx.VERTICAL)
            self.__toggleSizer = wx.BoxSizer(wx.HORIZONTAL)

        elif toggleSide in (wx.LEFT, wx.RIGHT):
            self.__mainSizer   = wx.BoxSizer(wx.HORIZONTAL)
            self.__toggleSizer = wx.BoxSizer(wx.VERTICAL)

        else:
            raise ValueError('toggleSide must be one of wx.TOP, '
                             'wx.BOTTOM, wx.LEFT or wx.RIGHT')

        if label is not None:
            label = six.u(label)

        self.__toggleButton = wx.StaticText(
            self,
            style=(wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_CENTRE_HORIZONTAL))

        if   toggleSide == wx.TOP:    hideLabel = six.u('\u25B2')
        elif toggleSide == wx.BOTTOM: hideLabel = six.u('\u25BC')
        elif toggleSide == wx.LEFT:   hideLabel = six.u('\u25C0')
        elif toggleSide == wx.RIGHT:  hideLabel = six.u('\u25B6')

        if   toggleSide == wx.TOP:    showLabel = six.u('\u25BC')
        elif toggleSide == wx.BOTTOM: showLabel = six.u('\u25B2')
        elif toggleSide == wx.LEFT:   showLabel = six.u('\u25B6')
        elif toggleSide == wx.RIGHT:  showLabel = six.u('\u25C0')

        self.__showLabel = showLabel
        self.__hideLabel = hideLabel
        self.__label     = label
        self.__state     = not initialState

        self.__toggleSizer.Add(self.__toggleButton, flag=wx.EXPAND)
        self.__toggleSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)

        if toggleSide in (wx.TOP, wx.LEFT):
            self.__mainSizer.Add(self.__toggleSizer,
                                 flag=wx.EXPAND | wx.ALL,
                                 border=2)
            self.__mainSizer.Add(self.__contentPanel,
                                 flag=wx.EXPAND,
                                 proportion=1)

        elif toggleSide in (wx.BOTTOM, wx.RIGHT):
            self.__mainSizer.Add(self.__contentPanel,
                                 flag=wx.EXPAND,
                                 proportion=1)
            self.__mainSizer.Add(self.__toggleSizer,
                                 flag=wx.EXPAND | wx.ALL,
                                 border=2)

        self.__toggleButton.Bind(wx.EVT_LEFT_DOWN, self.Toggle)

        self.SetSizer(self.__mainSizer)
        self.Expand(initialState)


    @property
    def button(self):
        """Returns the toggle button (actually a ``wx.StaticText``). This
        is for testing purposes.
        """
        return self.__toggleButton


    def __refresh(self):
        """Refreshes the layout. """
        self.Layout()
        self.__contentPanel.Layout()


    def __expand(self, expand=True, force=False):
        """Expands or collapses the content panel.

        :arg expand: ``True`` to expand, ``False`` to collapse.

        :arg force:  If ``True``, the panel is re-configured, regardless of
                     whether it is already in the requested expanded/collapsed
                     state.
        """

        if not force and self.IsExpanded() == expand:
            return

        self.__mainSizer.Show(self.__contentPanel, expand)

        if expand: label = self.__hideLabel
        else:      label = self.__showLabel

        if self.__label is not None:
            label = six.u('{} {}').format(label, self.__label)

        self.__state = expand
        self.__toggleButton.SetLabel(label)
        self.__refresh()


    def SetLabel(self, label):
        """Sets the label to show on the toggle button. Pass in ``None``
        for no label.
        """
        self.__label = label
        self.__expand(self.IsExpanded(), force=True)


    def GetLabel(self):
        """Returns the current toggle button label """
        return self.__label


    def GetToggleButton(self):
        """Returns the toggle button (actually a ``wx.StaticText``). This
        is for testing purposes.
        """
        warnings.warn('GetToggleButton is deprecated - use button instead',
                      category=DeprecationWarning,
                      stacklevel=2)

        return self.__toggleButton


    def Expand(self, expand=True):
        """Expand the content pane.

        :arg expand: ``True`` to expand, ``False`` to collapse.
        """
        self.__expand(expand)


    def Collapse(self):
        """Collapse the content pane. """
        self.Expand(False)


    def IsExpanded(self):
        """Return ``True`` if the content pane is currently expanded,
        ``False`` otherwise.
        """
        return self.__state


    def Toggle(self, ev=None):
        """Toggles visibility of the panel content.

        This method is called when the button is pushed. If ``ev`` is not
        ``None``, an ``EVT_TOGGLEPANEL_EVENT`` is generated.
        """

        newState = not self.IsExpanded()

        self.Expand(newState)

        if ev is not None:
            ev = TogglePanelEvent(newState=newState)
            ev.SetEventObject(self)
            wx.PostEvent(self, ev)


    def GetPane(self):
        """Returns the :class:`wx.Panel` to which all content should be
        added.
        """
        return self.__contentPanel


_TogglePanelEvent, _EVT_TOGGLEPANEL_EVENT = wxevent.NewEvent()


EVT_TOGGLEPANEL_EVENT = _EVT_TOGGLEPANEL_EVENT
"""Identifier for the :data:`TogglePanelEvent` event."""


TogglePanelEvent = _TogglePanelEvent
"""Event emitted when the toggle button is pushed. Contains the
following attributes:

  - ``newState``: The new visibility state of the toggle panel - ``True``
                  corresponds to visible, ``False`` to invisible.
"""



#!/usr/bin/env python
#
# test_textbitmap.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path   as op
import itertools as it

import matplotlib.image as mplimg

from . import compare_images

import fsleyes_widgets.utils.textbitmap as textbmp

datadir = op.join(op.dirname(__file__), 'testdata', 'textbitmap')


def test_textbitmap():

    texts     = ['R', 'Label']
    fontsizes = [6, 10, 16]
    bgColours = [(0, 0, 0, 0), (0, 0, 0, 1), (1, 0, 0, 1)]
    fgColours = [(0, 0, 0, 1), (1, 0, 0, 1), (1, 1, 1, 1)]
    alphas    = [0.5, 1.0]

    testcases = it.product(texts, fontsizes, bgColours, fgColours, alphas)

    for text, size, bg, fg, alpha in testcases:

        if bg == fg:
            continue

        bmp = textbmp.textBitmap(text, 75, 50, size, fg, bg, alpha)

        fname = [text, size] + list(bg) + list(fg) + [alpha]
        fname = '{}.png'.format('_'.join(map(str, fname)))
        fname = op.join(datadir, fname)

        benchmark = mplimg.imread(fname) * 255
        assert compare_images(bmp, benchmark, 0.1)[0]



#!/usr/bin/env python
#
# runwindow.py - Run a process, display its output in a wx window.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides classes and functions for running a non-interactive
process, and displaying its output.


This module provides the :class:`RunPanel` and :class:`ProcessManager`
classes, and a couple of associated convenience functions.

.. autosummary::
   :nosignatures:

   RunPanel
   ProcessManager
   run
"""

import os
import signal
import logging

import subprocess as sp
import threading

try:                import queue
except ImportError: import Queue as queue

import wx


log = logging.getLogger(__name__)


class RunPanel(wx.Panel):
    """A panel which displays a multiline text control, and a couple of
    buttons along the bottom. ``RunPanel`` instances are created by the
    :func:`run` function, and used/controlled by the :class:`ProcessManager`.


    One of the buttons is intended to closes the window in which this panel
    is contained. The second button is intended to terminate the running
    process. Both buttons are unbound by default, so must be manually
    configured by the creator.


    The text panel and buttons are available as the following attributes:

      - ``text``:        The text panel.
      - ``closeButton``: The `Close window` button.
      - ``killButton``:  The `Terminate process` button.
    """

    def __init__(self, parent):
        """Create a ``RunPanel``.

        :arg parent: The :mod:`wx` parent object.
        """
        wx.Panel.__init__(self, parent)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)

        # Horizontal scrolling no work in OSX
        # mavericks. I think it's a wxwidgets bug.
        self.text  = wx.TextCtrl(self,
                                 style=wx.TE_MULTILINE |
                                       wx.TE_READONLY  |
                                       wx.TE_DONTWRAP  |
                                       wx.HSCROLL)

        self.sizer.Add(self.text, flag=wx.EXPAND, proportion=1)

        self.btnPanel = wx.Panel(self)
        self.btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.btnPanel.SetSizer(self.btnSizer)

        self.sizer.Add(self.btnPanel, flag=wx.EXPAND)

        self.killButton  = wx.Button(self.btnPanel, label='Terminate process')
        self.closeButton = wx.Button(self.btnPanel, label='Close window')

        self.btnSizer.Add(self.killButton,  flag=wx.EXPAND, proportion=1)
        self.btnSizer.Add(self.closeButton, flag=wx.EXPAND, proportion=1)


class ProcessManager(threading.Thread):
    """A thread which manages the execution of a child process, and capture
    of its output.

    The process output is displayed in a :class:`RunPanel` which must be
    passed to the ``ProcessManager`` on creation.

    The :meth:`termProc` method can be used to terminate the child process
    before it has completed.
    """

    def __init__(self, cmd, parent, runPanel, onFinish):
        """Create a ``ProcessManager``.

        :arg cmd:      String or list of strings, the command to be
                       executed.

        :arg parent:   :mod:`wx` parent object.

        :arg runPanel: A :class:`RunPanel` instance , for displaying the
                       child process output.

        :arg onFinish: Callback function to be called when the process
                       finishes. May be ``None``. Must accept two parameters,
                       the GUI ``parent`` object, and the process return code.
        """
        threading.Thread.__init__(self, name=cmd[0])

        self.cmd      = cmd
        self.parent   = parent
        self.runPanel = runPanel
        self.onFinish = onFinish

        # Handle to the Popen object which represents
        # the child process. Created in run().
        self.proc = None

        # A queue for sharing data between the thread which
        # is blocking on process output (this thread object),
        # and the wx main thread which writes that output to
        # the runPanel
        self.outq = queue.Queue()

        # Put the command string at the top of the text control
        self.outq.put(' '.join(self.cmd) + '\n\n')
        wx.CallAfter(self.__writeToPanel)


    def __writeToPanel(self):
        """Reads a string from the output queue, and appends it
        to the :class:`RunPanel`. This method is intended to be
        executed via :func:`wx.CallAfter`.
        """

        try:                output = self.outq.get_nowait()
        except queue.Empty: output = None

        if output is None: return

        # ignore errors - the user may have closed the
        # runPanel window before the process has completed
        try:              self.runPanel.text.WriteText(output)
        except Exception: pass


    def run(self):
        """Starts the process, then reads its output line by line, writing
        each line asynchronously to the :class:`RunPanel`.  When the
        process ends, the ``onFinish`` method (if there is one) is called.
        If the process finishes abnormally (with a non-0 exit code) a warning
        dialog is displayed.
        """

        # Run the command. The preexec_fn parameter creates
        # a process group, so we are able to kill the child
        # process, and all of its children, if necessary.
        log.debug('Running process: "{}"'.format(' '.join(self.cmd)))
        self.proc = sp.Popen(self.cmd,
                             stdout=sp.PIPE,
                             bufsize=1,
                             stderr=sp.STDOUT,
                             preexec_fn=os.setsid)

        # read process output, line by line, pushing
        # each line onto the output queue and
        # asynchronously writing it to the runPanel
        for line in self.proc.stdout:

            log.debug('Process output: {}'.format(line.strip()))
            self.outq.put(line)
            wx.CallAfter(self.__writeToPanel)

        # When the above for loop ends, it means that the stdout
        # pipe has been broken. But it doesn't mean that the
        # subprocess is finished. So here, we wait until the
        # subprocess terminates, before continuing,
        self.proc.wait()

        retcode = self.proc.returncode

        log.debug(    'Process finished with return code {}'.format(retcode))
        self.outq.put('Process finished with return code {}'.format(retcode))

        wx.CallAfter(self.__writeToPanel)

        # Disable the 'terminate' button on the run panel
        def updateKillButton():

            # ignore errors - see __writeToPanel
            try:              self.runPanel.killButton.Enable(False)
            except Exception: pass

        wx.CallAfter(updateKillButton)

        # Run the onFinish handler, if there is one
        if self.onFinish is not None:
            wx.CallAfter(self.onFinish, self.parent, retcode)


    def termProc(self):
        """Attempts to kill the running child process."""

        log.debug('Attempting to send SIGTERM to '
                  'process group with pid {}'.format(self.proc.pid))
        os.killpg(self.proc.pid, signal.SIGTERM)

        # put a message on the runPanel
        self.outq.put('\nSIGTERM sent to process\n\n')
        wx.CallAfter(self.__writeToPanel)


def run(name, cmd, parent, onFinish=None, modal=True):
    """Runs the given command, displaying the output in a :class:`RunPanel`.

    :arg name:     Name of the tool to be run, used in the window title.

    :arg cmd:      String or list of strings, specifying the command to be
                   executed.

    :arg parent:   :mod:`wx` parent object.

    :arg modal:    If ``True``, the frame which contains the ``RunPanel``
                   will be modal.

    :arg onFinish: Function to be called when the process ends. Must
                   accept two parameters - a reference to the :mod:`wx`
                   frame/dialog displaying the process output, and
                   the exit code of the application.
    """

    # Create the GUI - if modal, the easiest
    # approach is to use a wx.Dialog
    if modal:
        frame = wx.Dialog(
            parent,
            title=name,
            style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
    else:
        frame = wx.Frame(parent, title=name)

    panel = RunPanel(frame)

    # Create the thread which runs the child process
    mgr = ProcessManager(cmd, parent, panel, onFinish)

    # Bind the panel control buttons so they do stuff
    panel.closeButton.Bind(wx.EVT_BUTTON, lambda e: frame.Close())
    panel.killButton .Bind(wx.EVT_BUTTON, lambda e: mgr.termProc())

    # Run the thread which runs the child process
    mgr.start()

    # layout and show the window
    frame.Layout()
    if modal: frame.ShowModal()
    else:     frame.Show()



#!/usr/bin/env python
#
# test_texttag.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import wx

from . import run_with_wx, simclick, simtext, simkey, addall, realYield

import fsleyes_widgets.texttag as tt


def test_StaticTextTag():
    run_with_wx(_test_StaticTextTag)
def _test_StaticTextTag():
    frame = wx.GetApp().GetTopWindow()
    tag   = tt.StaticTextTag(frame)
    dummy = wx.TextCtrl(frame)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(dummy, flag=wx.EXPAND)
    sizer.Add(tag,   flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    realYield()
    assert tag.GetText() == ''

    tag.SetText('TagText')
    tag.SetBackgroundColour('#bb3333')
    tag.SetBorderColour(    '#3333bb')
    realYield()
    assert tag.GetText() == 'TagText'

    dummy.SetFocus()
    realYield()
    tag.SetFocus()
    realYield()


def test_StaticTextTag_Close():
    run_with_wx(_test_StaticTextTag_Close)
def _test_StaticTextTag_Close():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    tag   = tt.StaticTextTag(frame)

    addall(frame, [tag])

    closeBtn = tag.closeButton

    called = [False]

    def handler(ev):
        called[0] = True

    tag.Bind(tt.EVT_STT_CLOSE, handler)

    tag.SetText('Blah')
    tag._StaticTextTag__onCloseButton(None)
    realYield()
    assert called[0]


def test_TextTagPanel_create_logic():
    run_with_wx(_test_TextTagPanel_create_logic)
def _test_TextTagPanel_create_logic():
    frame   = wx.GetApp().GetTopWindow()
    panel   = tt.TextTagPanel(frame)

    tags1    = ['Tag1',        'Tag2',        'Tag3']
    colours1 = [(50, 200, 50), (150, 20, 20), (50, 50, 225)]
    tags2    = ['2Tag1',        '2Tag2',         '3Tag3']
    colours2 = [(10, 100, 200), (100, 167, 202), (18, 28, 208)]

    panel.SetOptions(tags1, colours1)
    assert list(panel.GetOptions()) == tags1
    assert len(panel.GetTags()) == 0
    assert panel.TagCount() == 0
    for t, c in zip(tags1, colours1):
        assert tuple(panel.GetTagColour(t))         == c
        assert tuple(panel.GetTagColour(t.lower())) == c
    assert panel.GetTagColour(tags2[0]) is None

    # Replace the options with another set
    panel.SetOptions(tags2, colours2)
    assert list(panel.GetOptions()) == tags2
    assert len(panel.GetTags()) == 0
    assert panel.TagCount() == 0
    for t, c in zip(tags2, colours2):
        assert tuple(panel.GetTagColour(t))         == c
        assert tuple(panel.GetTagColour(t.lower())) == c

    # The ttp caches colours from old options
    # so colours1 should still be in there
    assert panel.GetTagColour(tags1[0]) == colours1[0]


def test_TextTagPanel_add_remove_logic():
    run_with_wx(_test_TextTagPanel_add_remove_logic)
def _test_TextTagPanel_add_remove_logic():
    frame   = wx.GetApp().GetTopWindow()
    panel   = tt.TextTagPanel(frame)

    tags    = ['Tag1',        'Tag2',        'Tag3']
    colours = [(50, 200, 50), (150, 20, 20), (50, 50, 225)]

    panel.SetOptions(tags, colours)

    panel.AddTag(tags[0])
    panel.AddTag(tags[1])

    assert panel.TagCount() == 2
    assert list(panel.GetTags()) == [tags[0], tags[1]]
    assert panel.GetTagIndex(tags[0]) == 0
    assert panel.GetTagIndex(tags[1]) == 1
    assert     panel.HasTag(tags[0])
    assert     panel.HasTag(tags[1])
    assert not panel.HasTag(tags[2])

    panel.SetTagColour(tags[0], (50, 50, 50))

    panel.RemoveTag(tags[0])
    assert panel.TagCount() == 1
    assert list(panel.GetTags()) == [tags[1]]
    assert not panel.HasTag(tags[0])
    assert     panel.HasTag(tags[1])

    panel.ClearTags()
    assert panel.TagCount() == 0
    assert list(panel.GetTags()) == []
    assert not panel.HasTag(tags[0])
    assert not panel.HasTag(tags[1])

    assert tuple(panel.GetTagColour(tags[0])) == (50, 50, 50)


def test_TextTagPanel_nostyle():
    run_with_wx(_test_TextTagPanel_nostyle)
def _test_TextTagPanel_nostyle():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = tt.TextTagPanel(frame, style=0)

    addall(frame, [panel])

    result = [None]

    def handler(ev):
        result[0] = ev.tag

    panel.Bind(tt.EVT_TTP_TAG_ADDED, handler)

    tags = ['Tag1', 'Tag2', 'Tag3']

    panel.SetOptions(tags)

    # Add an existing tag
    realYield()
    simtext(sim, panel.newTagCtrl.textCtrl, tags[0])
    assert panel.GetTags() == [tags[0]]
    assert result[0] == tags[0]

    simtext(sim, panel.newTagCtrl.textCtrl, tags[2])
    assert panel.GetTags() == [tags[0], tags[2]]
    assert result[0] == tags[2]

    # Duplicate
    result[0] = None
    simtext(sim, panel.newTagCtrl.textCtrl, tags[2])
    assert panel.GetTags() == [tags[0], tags[2], tags[2]]
    assert result[0] == tags[2]

    # Case insensitive
    simtext(sim, panel.newTagCtrl.textCtrl, tags[1].lower())
    assert panel.GetTags() == [tags[0], tags[2], tags[2], tags[1]]
    assert result[0] == tags[1]

    # Not in known tags
    result[0] = None
    simtext(sim, panel.newTagCtrl.textCtrl, 'notag')
    assert panel.GetTags() == [tags[0], tags[2], tags[2], tags[1]]
    assert result[0] is None


def test_TextTagPanel_close_event():
    run_with_wx(_test_TextTagPanel_close_event)
def _test_TextTagPanel_close_event():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=0)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    result = [None]

    def handler(ev):
        result[0] = ev.tag

    panel.Bind(tt.EVT_TTP_TAG_REMOVED, handler)

    panel.AddTag('MyTag1')
    panel.AddTag('MyTag2')
    realYield()

    panel.tags[1]._StaticTextTag__onCloseButton(None)
    realYield()
    assert result[0]       == 'MyTag2'
    assert panel.GetTags() == ['MyTag1']

    panel.tags[0]._StaticTextTag__onCloseButton(None)
    realYield()
    assert result[0]       == 'MyTag1'
    assert panel.GetTags() == []


def test_TextTagPanel_allow_new_tags():
    run_with_wx(_test_TextTagPanel_allow_new_tags)
def _test_TextTagPanel_allow_new_tags():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=tt.TTP_ALLOW_NEW_TAGS)

    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    realYield()
    simtext(sim, panel.newTagCtrl.textCtrl, 'MyNewTag')

    assert panel.GetTags()    == ['MyNewTag']
    assert panel.GetOptions() == []


def test_TextTagPanel_add_new_tags():
    run_with_wx(_test_TextTagPanel_add_new_tags)
def _test_TextTagPanel_add_new_tags():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=(tt.TTP_ALLOW_NEW_TAGS |
                                          tt.TTP_ADD_NEW_TAGS))

    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    realYield()
    simtext(sim, panel.newTagCtrl.textCtrl, 'MyNewTag')

    assert panel.GetTags()    == ['MyNewTag']
    assert panel.GetOptions() == ['MyNewTag']


def test_TextTagPanel_no_duplicates():
    run_with_wx(_test_TextTagPanel_no_duplicates)
def _test_TextTagPanel_no_duplicates():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=tt.TTP_NO_DUPLICATES)

    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    tags = ['Tag1', 'Tag2']
    panel.SetOptions(tags)
    realYield()
    simtext(sim, panel.newTagCtrl.textCtrl, tags[0])
    assert panel.GetTags() == [tags[0]]

    # Duplicate should not be added
    simtext(sim, panel.newTagCtrl.textCtrl, tags[0])
    assert panel.GetTags() == [tags[0]]

    simtext(sim, panel.newTagCtrl.textCtrl, tags[1])
    assert panel.GetTags() == [tags[0], tags[1]]


def test_TextTagPanel_case_sensitive():
    run_with_wx(_test_TextTagPanel_case_sensitive)
def _test_TextTagPanel_case_sensitive():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=(tt.TTP_CASE_SENSITIVE |
                                          tt.TTP_NO_DUPLICATES  |
                                          tt.TTP_ALLOW_NEW_TAGS))

    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    tags = ['Tag1', 'Tag2', 'Tag3', 'tag3']
    panel.SetOptions(tags)
    assert panel.GetOptions() == tags

    realYield()
    for i in range(len(tags)):
        simtext(sim, panel.newTagCtrl.textCtrl, tags[i])
        assert panel.GetTags() == tags[:i + 1]
        assert panel.HasTag(tags[i])

    panel.RemoveTag(tags[3])
    assert panel.GetTags() == tags[:3]

    panel.SetTagColour(tags[2], ( 50,  50,  50))
    panel.SetTagColour(tags[3], (100, 100, 100))

    assert panel.GetTagColour(tags[2]) == ( 50,  50,  50)
    assert panel.GetTagColour(tags[3]) == (100, 100, 100)


def test_TextTagPanel_mouse_focus():
    run_with_wx(_test_TextTagPanel_mouse_focus)
def _test_TextTagPanel_mouse_focus():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = tt.TextTagPanel(frame, style=tt.TTP_KEYBOARD_NAV)

    addall(frame, [panel])

    result = [None]

    def handler(ev):
        print('Gaaaargh')
        result[0] = ev.tag

    panel.Bind(tt.EVT_TTP_TAG_SELECT, handler)

    panel.SetOptions(['tag1', 'tag2'])
    panel.AddTag('tag1')

    simclick(sim, panel.tags[0], stype=2)
    assert result[0] == 'tag1'


def test_TextTagPanel_keyboard_nav():
    run_with_wx(_test_TextTagPanel_keyboard_nav)
def _test_TextTagPanel_keyboard_nav():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=tt.TTP_KEYBOARD_NAV)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    result = [None]

    def handler(ev):
        result[0] = ev.tag

    panel.Bind(tt.EVT_TTP_TAG_SELECT, handler)

    tags = ['tag1', 'tag2', 'tag3']

    panel.SetOptions(tags)
    for t in tags:
        panel.AddTag(t)
    realYield()

    panel.FocusNewTagCtrl()

    realYield()

    simkey(sim, panel.newTagCtrl, wx.WXK_LEFT)
    assert result[0] is None
    simkey(sim, panel.newTagCtrl, wx.WXK_RIGHT)
    assert result[0] == tags[0]

    simkey(sim, panel.tags[0], wx.WXK_RIGHT)
    assert result[0] == tags[1]
    simkey(sim, panel.tags[1], wx.WXK_RIGHT)
    assert result[0] == tags[2]
    result[0] = None
    simkey(sim, panel.tags[2], wx.WXK_RIGHT)
    assert result[0] is None

    simkey(sim, panel.tags[2], wx.WXK_LEFT)
    assert result[0] == tags[1]
    simkey(sim, panel.tags[1], wx.WXK_LEFT)
    assert result[0] == tags[0]
    result[0] = None
    simkey(sim, panel.tags[0], wx.WXK_LEFT)
    assert result[0] is None
    assert panel.newTagCtrl.textCtrl.HasFocus()



def test_TextTagPanel_keyboard_close():
    run_with_wx(_test_TextTagPanel_keyboard_close)
def _test_TextTagPanel_keyboard_close():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = tt.TextTagPanel(frame, style=tt.TTP_KEYBOARD_NAV)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    result = [None]

    def handler(ev):
        result[0] = ev.tag

    panel.Bind(tt.EVT_TTP_TAG_REMOVED, handler)

    tags = ['tag1', 'tag2', 'tag3']

    panel.SetOptions(tags)
    for t in tags:
        panel.AddTag(t)
    realYield()

    # Give a tag focus
    simclick(sim, panel.tags[0], stype=2)

    # Hit the delete key
    simkey(sim, panel.tags[0], wx.WXK_DELETE)
    assert result[0] == tags[0]

    # Give another tag focus
    simclick(sim, panel.tags[0], stype=2)
    # Hit the backspace key
    simkey(sim, panel.tags[0], wx.WXK_BACK)
    assert result[0] == tags[1]



#!/usr/bin/env python
#
# b64icon.py - convert base64-encoded bitmaps into wx.Bitmap obje
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :func:`loadBitmap` function, which can be used
to create a ``wx.Bitmap`` object from a base64-encoded image.
"""


import base64 as b64

from six import BytesIO

import wx


class IconError(Exception):
    """Custon ``Exception`` raised when :func:`loadBitmap` cannot load
    an icon.
    """
    pass


def loadBitmap(iconb64):
    """Convert the given ``base64``-encoded byte string to a ``wx.Bitmap``
    object.
    """

    iconbytes = b64.b64decode(iconb64)
    success   = False

    image   = wx.Image()
    success = wx.Image.LoadFile(image, BytesIO(iconbytes))

    if not success:
        raise IconError('Error loading image')

    return image.ConvertToBitmap()



#!/usr/bin/env python


import wx

import fsleyes_widgets.utils.overlay as overlay

from . import run_with_wx, realYield


def test_textOverlay():
    run_with_wx(_test_textOverlay)


def _test_textOverlay():
    frame = wx.GetTopLevelWindows()[0]
    overlay.textOverlay(frame, '1 2 3')



#!/usr/bin/env python
#
# typedict.py - Provides the TypeDict class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TypeDict` class, a type-aware dictionary.
"""


import six

from collections import abc


class TypeDict(object):
    """A type-aware dictionary.


    The purpose of the ``TypeDict`` is to allow value lookup using either
    classes or instances as keys. The ``TypeDict`` can be used in the same way
    that you would use a regular ``dict``, but the ``get`` and ``__getitem__``
    methods have some extra functionality.


    **Easy to understand example**


    Let's say we have a class with some properties::

        import fsleyes_widgets.utils.typedict as td

        class Animal(object):
            isMammal = True
            numLegs  = 4


    And we want to associate some tooltips with those properties::

        tooltips = td.TypeDict({

            'Animal.isMammal' : 'Set this to True for mammals, '
                                'False for reptiles.',
            'Animal.numLegs'  : 'The nuber of legs on this animal.'
        })

    Because we used a ``TypeDict``, we can now look up those tooltips
    in a number of ways::

        a = Animal()

        # Lookup by string (equivalent to a normal dict lookup)
        tt = tooltips['Animal.isMammal']

        # Lookup by class
        tt = tooltips[Animal, 'isMammal']

        # Lookup by instance
        tt = tooltips[a, 'isMammal']


    This functionality also works across class hierarchies::

        class Cat(Animal):
           numYoutubeHits = 10

        tooltips = td.TypeDict({

            'Animal.isMammal'    : 'Set this to True for mammals, '
                                   'False for reptiles.',
            'Animal.numLegs'     : 'The nuber of legs on this animal.',
            'Cat.numYoutubeHits' : 'Number of youtube videos this cat '
                                   'has starred in.'
        })

        c = Cat()

        isMammalTooltip    = tooltips[Cat,  'isMammal']
        numLegsTooltip     = tooltips[c,    'numLegs']
        youtubeHitsTooltip = tooltips[c,    'numYoutubeHits']

        # Class-hierachy-aware TypeDict lookups only
        # work when you pass in an instance/class as
        # the key - the following will result in a
        # KeyError:
        t = tooltips['Cat.numLegs']


    The :meth:`get` method has some extra functionality for working with
    class hierarchies::


        tooltips = td.TypeDict({

            'Animal.isMammal'    : 'Set this to True for mammals, '
                                   'False for reptiles.',

            'Animal.numLegs'     : 'The nuber of legs on this animal.',

            'Cat.numLegs'        : 'This will be equal to four for all cats, '
                                    'but could be less for disabled cats, '
                                    'or more for lucky cats.',

            'Cat.numYoutubeHits' : 'Number of youtube videos this cat '
                                   'has starred in.'
        })

        print tooltips.get((c, 'numLegs'))
        #  'This will be equal to four for all cats, but could '
        #  'be less for disabled cats, or more for lucky cats.'

        print tooltips.get((c, 'numLegs'), allhits=True)
        # ['This will be equal to four for all cats, but could '
        #  'be less for disabled cats, or more for lucky cats.',
        #  'The nuber of legs on this animal.']

        print tooltips.get((c, 'numLegs'), allhits=True, bykey=True)
        # {('Animal', 'numLegs'): 'The nuber of legs on this animal.',
        #  ('Cat',    'numLegs'): 'This will be equal to four for all cats, '
        #                         'but could be less for disabled cats, or '
        #                         'more for lucky cats.'}


    **Boring technical description**


    The ``TypeDict`` is a custom dictionary which allows classes or class
    instances to be used as keys for value lookups, but internally transforms
    any class/instance keys into strings. Tuple keys are supported. Value
    assignment with class/instance keys is not supported. All keys are
    transformed via the :meth:`tokenifyKey` method before being internally
    used and/or stored.

    If a class/instance is passed in as a key, and there is no value
    associated with that class, a search is performed on all of the base
    classes of that class to see if any values are present for them.
    """


    def __init__(self, initial=None):
        """Create a ``TypeDict``.

        :arg initial: Dictionary containing initial values.
        """

        if initial is None:
            initial = {}

        self.__dict = {}

        for k, v in initial.items():
            self[k] = v


    def __str__(self):
        return self.__dict.__str__()


    def __repr__(self):
        return self.__dict.__repr__()


    def __len__(self):
        return len(self.__dict)


    def keys(self):
        return self.__dict.keys()


    def values(self):
        return self.__dict.values()


    def items(self):
        return self.__dict.items()


    def __setitem__(self, key, value):
        self.__dict[self.tokenifyKey(key)] = value


    def tokenifyKey(self, key):
        """Turns a dictionary key, which may have been specified as a
        string, or a combination of strings and types, into a tuple.
        """

        if isinstance(key, six.string_types):
            if '.' in key: return tuple(key.split('.'))
            else:          return key

        if isinstance(key, abc.Sequence):

            tKeys = map(self.tokenifyKey, key)
            key   = []

            for tk in tKeys:
                if   isinstance(tk, six.string_types): key.append(tk)
                elif isinstance(tk, abc.Sequence):     key += list(tk)
                else:                                  key.append(tk)

            return tuple(key)

        return key


    def get(self, key, default=None, allhits=False, bykey=False, exact=False):
        """Retrieve the value associated with the given key. If
        no value is present, return the specified ``default`` value,
        which itself defaults to ``None``.

        If the specified key contains a class or instance, and the ``exact``
        argument is ``False`` (the default), the entire class hierarchy is
        searched, and the first value present for the class, or any base
        class, are returned. If ``exact is True`` and no value exists
        for the specific class, the ``default`` is returned.

        If ``exact is False`` and the ``allhits`` argument evaluates to
        ``True``, the entire class hierarchy is searched, and all values
        present for the class, and any base class, are returned as a sequence.

        If ``allhits`` is ``True`` and the ``bykey`` parameter is also
        set to ``True``, a dictionary is returned rather than a sequence,
        where the dictionary contents are the subset of this dictionary,
        containing the keys which equated to the given key, and their
        corresponding values.
        """

        try:             return self.__getitem__(key, allhits, bykey, exact)
        except KeyError: return default


    def __getitem__(self, key, allhits=False, bykey=False, exact=False):

        origKey = key
        key     = self.tokenifyKey(key)
        bases   = []

        # Make the code a bit easier by
        # treating non-tuple keys as tuples
        if not isinstance(key, tuple):
            key = tuple([key])

        newKey = []

        # Transform any class/instance elements into
        # their string representation (the class name)
        for elem in key:

            if isinstance(elem, type):
                newKey.append(elem.__name__)
                bases .append(elem.__bases__)

            elif not isinstance(elem, (str, int)):
                newKey.append(elem.__class__.__name__)
                bases .append(elem.__class__.__bases__)

            else:
                newKey.append(elem)
                bases .append(None)

        if exact:
            bases = []

        key  = newKey

        keys = []
        hits = []

        while True:

            # If the key was not a tuple turn
            # it back into a single element key
            # for the lookup
            if len(key) == 1: lKey = key[0]
            else:             lKey = tuple(key)

            val = self.__dict.get(lKey, None)

            # We've found a value for the key
            if val is not None:

                # If allhits is false, just return the value
                if not allhits: return val

                # Otherwise, accumulate the value, and keep
                # searching
                else:
                    hits.append(val)
                    if bykey:
                        keys.append(lKey)

            # No more base classes to search for - there
            # really is no value associated with this key
            elif all([b is None for b in bases]):
                raise KeyError(key)

            # Search through the base classes to see
            # if a value is present for one of them
            for i, (elem, elemBases) in enumerate(zip(key, bases)):
                if elemBases is None:
                    continue

                # test each of the base classes
                # of the current tuple element
                for elemBase in elemBases:

                    newKey    = list(key)
                    newKey[i] = elemBase

                    if len(newKey) == 1: newKey = newKey[0]
                    else:                newKey = tuple(newKey)

                    try:
                        newVal = self.__getitem__(newKey, allhits, bykey)
                    except KeyError:
                        continue

                    if not allhits:
                        return newVal
                    else:
                        if bykey:
                            newKeys, newVals = zip(*newVal.items())
                            keys.extend(newKeys)
                            hits.extend(newVals)
                        else:
                            hits.extend(newVal)

            # No value for any base classes either
            if len(hits) == 0:
                raise KeyError(origKey)

            # if bykey is true, return a dict
            # containing the values and their
            # corresponding keys
            if bykey:
                return dict(zip(keys, hits))

            # otherwise just return the
            # list of matched values
            else:
                return hits



#!/usr/bin/env python
#
# dialog.py - Miscellaneous dialogs.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains a collection of basic dialog classes, available for
use throughout ``fslpy``. The available dialogs are:

.. autosummary::
   :nosignatures:

   SimpleMessageDialog
   TimeoutDialog
   ProcessingDialog
   TextEditDialog
   FSLDirDialog
"""


import            os
import os.path as op
import            threading

import            six
import            wx

import fsleyes_widgets as fw


class SimpleMessageDialog(wx.Dialog):
    """A simple, no-frills :class:`wx.Dialog` for displaying a message. The
    message can be updated via the :meth:`SetMessage` method. As a simple
    usage example::

        import fsleyes_widgets.dialog as fsldlg
        dlg = fsldlg.SimpleMessageDialog(message='Loading data ...')

        dlg.Show()

        # load the data, like
        # you said you would

        # Data is loaded, so we
        # can kill the dialog
        dlg.Close()
        dlg.Destroy()


    The ``SimpleMessageDialog`` class supports the following styles:

    .. autosummary::
       SMD_KEEP_CENTERED


    a ``SimpleMessageDialog`` looks something like this:

    .. image:: images/simplemessagedialog.png
       :scale: 50%
       :align: center
    """


    def __init__(self, parent=None, message='', style=None):
        """Create a ``SimpleMessageDialog``.

        :arg parent:  The :mod:`wx` parent object.

        :arg message: The initial message to show.

        :arg style:   Only one style flag  is supported,
                      :data:`SMD_KEEP_CENTERED`. This flag is enabled by
                      default.
        """


        if style is None:
            style = SMD_KEEP_CENTERED

        if parent is None:
            parent = wx.GetApp().GetTopWindow()

        wx.Dialog.__init__(self,
                           parent,
                           style=wx.STAY_ON_TOP | wx.FULL_REPAINT_ON_RESIZE)


        self.__style = style

        self.__message = wx.StaticText(
            self,
            style=(wx.ST_ELLIPSIZE_MIDDLE     |
                   wx.ALIGN_CENTRE_HORIZONTAL |
                   wx.ALIGN_CENTRE_VERTICAL))

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__message,
                         border=25,
                         proportion=1,
                         flag=wx.EXPAND | wx.CENTRE | wx.ALL)

        self.SetTransparent(240)
        self.SetBackgroundColour((225, 225, 255))

        self.SetSizer(self.__sizer)

        self.SetMessage(message)


    def Show(self):
        """Overrides ``wx.Dialog.Show``. Calls that method, and calls
        ``wx.Yield``.
        """
        wx.Dialog.Show(self)
        wx.Yield()


    def SetMessage(self, msg):
        """Updates the message shown on this ``SimpleMessageDialog``.

        If the :data:`SMD_KEEP_CENTERED` style is set, the dialog is
        re-centered on its parent, to account for changes in the message width.
        """

        msg = str(msg)

        self.__message.SetLabel(msg)

        # Figure out the dialog size
        # required to fit the message
        dc = wx.ClientDC(self.__message)

        width, height = dc.GetTextExtent(msg)

        # +50 to account for sizer borders (see __init__),
        # plus a bit more for good measure. In particular,
        # under GTK, the message seems to be vertically
        # truncated if we don't add some extra padding
        width  += 60
        height += 70

        self.SetMinClientSize((width, height))
        self.SetClientSize((   width, height))

        self.Layout()
        self.__message.Layout()

        if self.__style & SMD_KEEP_CENTERED:
            self.CentreOnParent()

        # This ridiculousness seems to be
        # necessary to force a repaint on
        # all platforms (OSX, GTK, GTK/SSH)
        wx.Yield()
        self.Refresh()
        self.Update()
        self.__message.Refresh()
        self.__message.Update()
        wx.Yield()


# SimpleMessageDialog style flags
SMD_KEEP_CENTERED = 1
"""If set, the dialog will be re-centred on its parent whenever its message
changes.
"""


class TimeoutDialog(SimpleMessageDialog):
    """A :class:`SimpleMessageDialog` which automatically destroys itself
    after a specified timeout period.

     .. note:: The timeout functionality will not work if you show the dialog
               by any means other than the :meth:`wx.Dialog.Show` or
               :meth:`wx.Dialog.ShowModal` methods ... but is there any other
               way of showing a :class:`wx.Dialog`?
    """


    def __init__(self, parent, message, timeout=1000, **kwargs):
        """Create a ``TimeoutDialog``.

        :arg parent:  The :mod:`wx` parent object.

        :arg message: The initial message to display.

        :arg timeout: Timeout period in milliseconds.

        :arg kwargs:  Passed through to :meth:`SimpleMessageDialog.__init__`.
        """

        SimpleMessageDialog.__init__(self, parent, message, **kwargs)
        self.__timeout = timeout


    def __close(self):
        """Closes and destroys this ``TimeoutDialog``. """
        self.Close()
        self.Destroy()


    def Show(self):
        """Shows this ``TimeoutDialog``, and sets up a callback to
        close it after the specified ``timeout``.
        """
        wx.CallLater(self.__timeout, self.__close)
        SimpleMessageDialog.Show(self)


    def ShowModal(self):
        """Shows this ``TimeoutDialog``, and sets up a callback to
        close it after the specified ``timeout``.
        """
        wx.CallLater(self.__timeout, self.__close)
        SimpleMessageDialog.ShowModal(self)


class ProcessingDialog(SimpleMessageDialog):
    """A :class:`SimpleMessageDialog` which displays a message and runs a
    task in the background. User interaction is blocked while the task runs,
    and the dialog closes and destroys itself automatically on task
    completion.


    The task is simply passed in as a function. If the task supports it,
    the ``ProcessingDialog`` will pass it two message-updating functions,
    which can be used by the task to update the message being displayed.
    This functionality is controlled by the ``passFuncs``, ``messageFunc``
    and ``errorFunc`` parameters to :meth:`__init__`.


    A ``ProcessingDialog`` must be displayed via the :meth:`Run` method,
    *not* with the :meth:`wx.Dialog.Show` or :meth:`wx.Dialog.ShowModal`
    methods.
    """

    def __init__(self, parent, message, task, *args, **kwargs):
        """Create a ``ProcessingDialog``.

        :arg parent:       The :mod:`wx` parent object.

        :arg message:      Initial message to display.

        :arg task:         The function to run.

        :arg args:         Positional arguments passed to the ``task``
                           function.

        :arg kwargs:       Keyword arguments passed to the ``task`` function.


        Some special keyword arguments are also accepted:

        ===============  =================================================
        Name             Description
        ===============  =================================================
        ``passFuncs``    If ``True``, two extra keyword arguments  are
                         passed to the ``task`` function - ``messageFunc``
                         and ``errorFunc``.

                         ``messageFunc`` is a function which accepts a
                         single string as its argument; when it is called,
                         the dialog  message is updated to display the
                         string.

                         ``errorFunc`` is a function which accepts two
                         arguemnts - a message string and an
                         :exc:`Exception` instance. If the task detects
                         an error, it may call this function. A new
                         dialog is shown, containing the details of the
                         error, to inform the user.
        ``messageFunc``  Overrides the default ``messageFunc`` described
                         above.
        ``errorFunc``    Overrides the default ``errorFunc`` described
                         above.
        ===============  =================================================
        """

        passFuncs = kwargs.get('passFuncs', False)

        if not passFuncs:
            kwargs.pop('messageFunc', None)
            kwargs.pop('errorFunc',   None)
        else:
            kwargs['messageFunc'] = kwargs.get('messageFunc',
                                               self.__defaultMessageFunc)
            kwargs['errortFunc']  = kwargs.get('errorFunc',
                                               self.__defaultErrorFunc)

        self.task    = task
        self.args    = args
        self.kwargs  = kwargs
        self.message = message

        style = kwargs.pop('style', None)

        SimpleMessageDialog.__init__(self, parent, style=style)


    def Run(self, mainThread=False):
        """Shows this ``ProcessingDialog``, and runs the ``task`` function
        passed to :meth:`__init__`. When the task completes, this dialog
        is closed and destroyed.

        :arg mainThread: If ``True`` the task is run in the current thread.
                         Otherwise, the default behaviour is to run the
                         task in a separate thread.

        :returns: the return value of the ``task`` function.

        .. note:: If ``mainThread=True``, the task should call
                  :func:`wx.Yield` periodically - under GTK, there is a
                  chance that this ``ProcessingDialog`` will not get drawn
                  before the task begins.
        """

        self.SetMessage(self.message)
        wx.Dialog.Show(self)
        self.SetFocus()

        self.Refresh()
        self.Update()
        wx.Yield()

        if mainThread:
            try:
                result = self.task(*self.args, **self.kwargs)
            except:
                self.Close()
                self.Destroy()
                raise
        else:
            returnVal = [None]

            def wrappedTask():
                returnVal[0] = self.task(*self.args, **self.kwargs)

            thread = threading.Thread(target=wrappedTask)
            thread.start()

            while thread.isAlive():
                thread.join(0.2)
                wx.Yield()

            result = returnVal[0]

        self.Close()
        self.Destroy()

        return result


    def Show(self):
        """Raises a :exc:`NotImplementedError`."""
        raise NotImplementedError('Use the Run method')


    def ShowModal(self):
        """Raises a :exc:`NotImplementedError`."""
        raise NotImplementedError('Use the Run method')


    def __defaultMessageFunc(self, msg):
        """Default ``messageFunc``. Updates the message which is displayed
        on this ``ProcessingDialog``. See :meth:`SetMessage`.
        """
        self.SetMessage(msg)


    def __defaultErrorFunc(self, msg, err):
        """Default ``errorFunc``. Opens a new dialog (a :class:`wx.MessageBox`)
        which contains a description of the error.
        """
        err   = str(err)
        msg   = 'An error hass occurred: {}\n\nDetails: {}'.format(msg, err)
        title = 'Error'
        wx.MessageBox(msg, title, wx.ICON_ERROR | wx.OK)


class TextEditDialog(wx.Dialog):
    """A dialog which shows an editable/selectable text field.


    ``TextEditDialog`` supports the following styles:

    .. autosummary::
       TED_READONLY
       TED_MULTILINE
       TED_OK
       TED_CANCEL
       TED_OK_CANCEL
       TED_COPY
       TED_COPY_MESSAGE

    A ``TextEditDialog`` looks something like this:

    .. image:: images/texteditdialog.png
       :scale: 50%
       :align: center
    """

    def __init__(self,
                 parent,
                 title='',
                 message='',
                 text='',
                 icon=None,
                 style=None):
        """Create a ``TextEditDialog``.

        :arg parent:  The :mod:`wx` parent object.

        :arg title:   Dialog title.

        :arg message: Dialog message.

        :arg text:    String  to display in the text field.

        :arg icon:    A :mod:`wx` icon identifier, such as
                      :data:`wx.ICON_INFORMATION` or :data:`wx.ICON_WARNING`.

        :arg style:   A combination of :data:`TED_READONLY`,
                      :data:`TED_MULTILINE`, :data:`TED_OK`,
                      :data:`TED_CANCEL`, :data:`TED_OK_CANCEL`,
                      :data:`TED_COPY` and :data:`TED_COPY_MESSAGE` . Defaults
                      to :data:`TED_OK`.
        """

        if style is None:
            style = TED_OK

        wx.Dialog.__init__(self,
                           parent,
                           title=title,
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        textStyle = 0
        if style & TED_READONLY:  textStyle |= wx.TE_READONLY
        if style & TED_MULTILINE: textStyle |= wx.TE_MULTILINE

        self.__message  = wx.StaticText(self)
        self.__textEdit = wx.TextCtrl(  self, style=textStyle)

        self.__message .SetLabel(message)
        self.__textEdit.SetValue(text)

        self.__showCopyMessage = style & TED_COPY_MESSAGE

        # set the min size of the text
        # ctrl so it can fit a few lines
        self.__textEdit.SetMinSize((-1, 120))
        self.__textEdit.SetMaxSize((600, -1))

        self.__ok      = (-1, -1)
        self.__copy    = (-1, -1)
        self.__cancel  = (-1, -1)
        self.__icon    = (-1, -1)
        self.__buttons = []

        if icon is not None:

            icon = wx.ArtProvider.GetMessageBoxIcon(icon)

            if fw.wxversion() == fw.WX_PHOENIX:
                bmp = wx.Bitmap()
            else:
                bmp = wx.EmptyBitmap(icon.GetWidth(), icon.GetHeight())

            bmp.CopyFromIcon(icon)
            self.__icon = wx.StaticBitmap(self)
            self.__icon.SetBitmap(bmp)

        if style & TED_OK:
            self.__ok = wx.Button(self, id=wx.ID_OK)
            self.__ok.Bind(wx.EVT_BUTTON, self.onOk)
            self.__buttons.append(self.__ok)

        if style & TED_CANCEL:
            self.__cancel = wx.Button(self, id=wx.ID_CANCEL)
            self.__cancel.Bind(wx.EVT_BUTTON, self.onCancel)
            self.__buttons.append(self.__cancel)

        if style & TED_COPY:
            self.__copy = wx.Button(self, label='Copy to clipboard')
            self.__copy.Bind(wx.EVT_BUTTON, self.__onCopy)
            self.__buttons.append(self.__copy)

        self.__textEdit.Bind(wx.EVT_CHAR_HOOK, self.__onCharHook)

        textSizer = wx.BoxSizer(wx.VERTICAL)
        iconSizer = wx.BoxSizer(wx.HORIZONTAL)
        btnSizer  = wx.BoxSizer(wx.HORIZONTAL)
        mainSizer = wx.BoxSizer(wx.VERTICAL)

        textSizer.Add(self.__message,
                      flag=wx.ALL | wx.CENTRE,
                      border=20)
        textSizer.Add(self.__textEdit,
                      flag=wx.ALL | wx.EXPAND,
                      border=20,
                      proportion=1)

        iconSizer.Add(self.__icon, flag=wx.ALL | wx.CENTRE, border=20)
        iconSizer.Add(textSizer, flag=wx.EXPAND, proportion=1)

        btnSizer.AddStretchSpacer()
        btnSizer.Add(self.__ok,
                     flag=wx.ALL | wx.CENTRE,
                     border=10)
        btnSizer.Add(self.__copy,
                     flag=wx.ALL | wx.CENTRE,
                     border=10)
        btnSizer.Add(self.__cancel,
                     flag=wx.ALL | wx.CENTRE,
                     border=10)
        btnSizer.Add((-1, 20))

        mainSizer.Add(iconSizer, flag=wx.EXPAND, proportion=1)
        mainSizer.Add(btnSizer,  flag=wx.EXPAND)

        self.SetSizer(mainSizer)
        self.Fit()


    def __onCharHook(self, ev):
        """Called on ``EVT_CHAR_HOOK`` events generated by the ``TextCtrl``.
        Implements tab-navigation, and makes the enter key behave as if
        the user had clicked the OK button.
        """

        key = ev.GetKeyCode()

        if key not in (wx.WXK_TAB, wx.WXK_RETURN):
            ev.Skip()
            return

        # Dodgy, but I've had loads of trouble
        # under OSX - Navigate/HandleAsNavigationKey
        # do not work.
        if key == wx.WXK_TAB and len(self.__buttons) > 0:
            self.__buttons[0].SetFocus()

        elif key == wx.WXK_RETURN:
            self.onOk(None)


    def onOk(self, ev):
        """Called when the *Ok* button is pressed. Ends the dialog. """
        self.EndModal(wx.ID_OK)


    def onCancel(self, ev):
        """Called when the *Cancel* button is pressed. Ends the dialog. """
        self.EndModal(wx.ID_CANCEL)


    def __onCopy(self, ev):
        """Called when the *Copy* button is pressed. Copies the text
        to the system clipboard, and pops up a :class:`TimeoutDialog`
        informing the user.
        """
        text = self.__textEdit.GetValue()

        cb = wx.TheClipboard

        if cb.Open():
            cb.SetData(wx.TextDataObject(text))
            cb.Close()

            if self.__showCopyMessage:
                td = TimeoutDialog(self, 'Copied!', 1000)
                td.Show()


    def SetMessage(self, message):
        """Set the message displayed on the dialog."""
        self.__message.SetLabel(message)


    def SetOkLabel(self, label):
        """Set the label to show on the *Ok* button."""
        self.__ok.SetLabel(label)


    def SetCopyLabel(self, label):
        """Sets the label to show on the *Copy* button."""
        self.__copy.SetLabel(label)


    def SetCancelLabel(self, label):
        """Sets the label to show on the *Cancel* button."""
        self.__cancel.SetLabel(label)


    def SetText(self, text):
        """Sets the text to show in the text field."""
        self.__textEdit.SetValue(text)


    def GetText(self):
        """Returns the text shown in the text field."""
        return self.__textEdit.GetValue()


# TextEditDialog style flags


TED_READONLY = 1
"""If set, the user will not be able to change the text field contents."""


TED_MULTILINE = 2
"""If set, the text field will span multiple lines. """


TED_OK = 4
"""If set, an *Ok* button will be shown. """


TED_CANCEL = 8
"""If set, a *Cancel* button will be shown. """


TED_OK_CANCEL = 12
"""If set, *Ok* and *Cancel* buttons will be shown. Equivalent to
``TED_OK | TED_CANCEL``.
"""


TED_COPY = 16
"""If set, a *Copy* button will be shown, allowing the use to copy
the text to the system clipboard.
"""


TED_COPY_MESSAGE = 32
"""If set, and if :attr:`TED_COPY` is also set, when the user chooses
to copy the text to the system clipboard, a popup message is displayed.
"""


class FSLDirDialog(wx.Dialog):
    """A dialog which warns the user that the ``$FSLDIR`` environment
    variable is not set, and prompts them to identify the FSL
    installation directory.

    If the user selects a directory, the :meth:`getFSLDir` method can be
    called to retrieve their selection after the dialog has been closed.

    A ``FSLDirDialog`` looks something like this:

    .. image:: images/fsldirdialog.png
       :scale: 50%
       :align: center
    """

    def __init__(self, parent, toolName, osxHint):
        """Create a ``FSLDirDialog``.

        :arg parent:   The :mod:`wx` parent object.

        :arg toolName: The name of the tool which is running.

        :arg osxHint:  If ``True``, an OSX-specific hint is added to the
                       dialog.
        """

        wx.Dialog.__init__(self, parent, title='$FSLDIR is not set')

        self.__fsldir  = None
        self.__icon    = wx.StaticBitmap(self)
        self.__message = wx.StaticText(  self)
        self.__locate  = wx.Button(      self, id=wx.ID_OK)
        self.__skip    = wx.Button(      self, id=wx.ID_CANCEL)

        icon = wx.ArtProvider.GetMessageBoxIcon(wx.ICON_EXCLAMATION)

        if fw.wxversion() == fw.WX_PHOENIX:
            bmp = wx.Bitmap()
        else:
            bmp  = wx.EmptyBitmap(icon.GetWidth(), icon.GetHeight())

        bmp.CopyFromIcon(icon)

        self.__icon.SetBitmap(bmp)
        self.__message.SetLabel(
            'The $FSLDIR environment variable is not set - {} '
            'may not behave correctly.'.format(toolName))
        self.__locate .SetLabel('Locate $FSLDIR')
        self.__skip   .SetLabel('Skip')

        self.__skip  .Bind(wx.EVT_BUTTON, self.__onSkip)
        self.__locate.Bind(wx.EVT_BUTTON, self.__onLocate)

        self.__mainSizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__contentSizer = wx.BoxSizer(wx.VERTICAL)
        self.__buttonSizer  = wx.BoxSizer(wx.HORIZONTAL)

        self.__buttonSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)
        self.__buttonSizer.Add(self.__locate)
        self.__buttonSizer.Add((20, 1))
        self.__buttonSizer.Add(self.__skip)

        self.__contentSizer.Add(self.__message, flag=wx.EXPAND, proportion=1)
        self.__contentSizer.Add((1, 20))
        self.__contentSizer.Add(self.__buttonSizer, flag=wx.EXPAND)

        # If running on OSX, add a message
        # telling the user about the
        # cmd+shift+g shortcut
        if osxHint:

            self.__hint = wx.StaticText(
                self,
                label=six.u('Hint: Press \u2318+\u21e7+G in the file '
                            'dialog to manually type in a location.'))

            self.__hint.SetForegroundColour('#888888')

            self.__contentSizer.Insert(2, self.__hint, flag=wx.EXPAND)
            self.__contentSizer.Insert(3, (1, 20))

        else:
            self.__hint = None

        self.__mainSizer.Add(self.__icon,
                             flag=wx.ALL | wx.CENTRE,
                             border=20)
        self.__mainSizer.Add(self.__contentSizer,
                             flag=wx.EXPAND | wx.ALL,
                             proportion=1,
                             border=20)

        self.__message.Wrap(self.GetSize().GetWidth())

        self.SetSizer(self.__mainSizer)
        self.__mainSizer.Layout()
        self.__mainSizer.Fit(self)

        self.CentreOnParent()


    def GetFSLDir(self):
        """If the user selected a directory, this method returns their
        selection. Otherwise, it returns ``None``.
        """
        return self.__fsldir


    def __onSkip(self, ev):
        """called when the *Skip* button is pushed. """
        self.EndModal(wx.ID_CANCEL)


    def __onLocate(self, ev):
        """Called when the *Locate* button is pushed. Opens a
        :class:`wx.DirDialog` which allows the user to locate the
        FSL installation directory.
        """

        dlg = wx.DirDialog(
            self,
            message='Select the directory in which FSL is installed',
            defaultPath=op.join(os.sep, 'usr', 'local'),
            style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)

        # If the user cancels the file
        # dialog, focus returns to the
        # original 'Choose dir' / 'Skip'
        # dialog.
        if dlg.ShowModal() != wx.ID_OK:
            return

        self.__fsldir = dlg.GetPath()

        self.EndModal(wx.ID_OK)


class CheckBoxMessageDialog(wx.Dialog):
    """A ``wx.Dialog`` which displays a message, one or more ``wx.CheckBox``
    widgets, with associated messages, an *Ok* button, and (optionally) a
    *Cancel* button.
    """


    def __init__(self,
                 parent,
                 title=None,
                 message=None,
                 cbMessages=None,
                 cbStates=None,
                 yesText=None,
                 noText=None,
                 cancelText=None,
                 hintText=None,
                 focus=None,
                 icon=None,
                 style=None):
        """Create a ``CheckBoxMessageDialog``.

        :arg parent:        A ``wx`` parent object.

        :arg title:         The dialog frame title.

        :arg message:       Message to show on the dialog.

        :arg cbMessages:    A list of labels, one for each ``wx.CheckBox``.

        :arg cbStates:      A list of initial states (boolean values) for
                            each ``wx.CheckBox``.

        :arg yesText:       Text to show on the *yes*/confirm button. Defaults
                            to *OK*.

        :arg noText:        Text to show on the *no* button. If not provided,
                            there will be no *no* button.

        :arg cancelText:    Text to show on the *cancel* button. If not
                            provided, there will be no cancel button.

        :arg hintText:      If provided, shown as a "hint", in a slightly
                            faded font, between the checkboxes and the buttons.

        :arg focus:         One of ``'yes'``, ``'no'```, or ``'cancel'``,
                            specifying which button should be given initial
                            focus.

        :arg icon:          A ``wx`` icon identifier (e.g.
                            ``wx.ICON_EXCLAMATION``).

        :arg style:         Passed through to the ``wx.Dialog.__init__``
                            method. Defaults to ``wx.DEFAULT_DIALOG_STYLE``.
        """

        if style      is None: style      = wx.DEFAULT_DIALOG_STYLE
        if title      is None: title      = ''
        if message    is None: message    = ''
        if cbMessages is None: cbMessages = ['']
        if cbStates   is None: cbStates   = [False] * len(cbMessages)
        if yesText    is None: yesText    = 'OK'

        wx.Dialog.__init__(self, parent, title=title, style=style)

        if icon is not None:
            icon = wx.ArtProvider.GetMessageBoxIcon(icon)
            self.__icon = wx.StaticBitmap(self)

            if fw.wxversion() == fw.WX_PHOENIX:
                bmp = wx.Bitmap()
            else:
                bmp = wx.EmptyBitmap(icon.GetWidth(), icon.GetHeight())

            bmp.CopyFromIcon(icon)
            self.__icon.SetBitmap(bmp)
        else:
            self.__icon = (1, 1)

        self.__checkboxes = []
        for msg, state in zip(cbMessages, cbStates):
            cb = wx.CheckBox(self, label=msg)
            cb.SetValue(state)
            self.__checkboxes.append(cb)

        self.__message   = wx.StaticText(self, label=message)
        self.__yesButton = wx.Button(    self, label=yesText, id=wx.ID_YES)

        self.__yesButton.Bind(wx.EVT_BUTTON, self.__onYesButton)

        if noText is not None:
            self.__noButton = wx.Button(self, label=noText, id=wx.ID_NO)
            self.__noButton.Bind(wx.EVT_BUTTON, self.__onNoButton)

        else:
            self.__noButton = None

        if cancelText is not None:
            self.__cancelButton = wx.Button(self,
                                            label=cancelText,
                                            id=wx.ID_CANCEL)
            self.__cancelButton.Bind(wx.EVT_BUTTON, self.__onCancelButton)
        else:
            self.__cancelButton = None

        if hintText is not None:
            self.__hint = wx.StaticText(self, label=hintText)
            self.__hint.SetForegroundColour('#888888')
        else:
            self.__hint = None

        self.__mainSizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__contentSizer = wx.BoxSizer(wx.VERTICAL)
        self.__btnSizer     = wx.BoxSizer(wx.HORIZONTAL)

        self.__contentSizer.Add(self.__message,  flag=wx.EXPAND, proportion=1)
        self.__contentSizer.Add((1, 20), flag=wx.EXPAND)
        for cb in self.__checkboxes:
            self.__contentSizer.Add(cb, flag=wx.EXPAND)

        if self.__hint is not None:
            self.__contentSizer.Add((1, 20), flag=wx.EXPAND)
            self.__contentSizer.Add(self.__hint, flag=wx.EXPAND)

        self.__contentSizer.Add((1, 20), flag=wx.EXPAND)
        self.__btnSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)

        buttons = [self.__yesButton, self.__noButton, self.__cancelButton]
        buttons = [b for b in buttons if b is not None]

        for i, b in enumerate(buttons):
            self.__btnSizer.Add(b)
            if i != len(buttons) - 1:
                self.__btnSizer.Add((5, 1), flag=wx.EXPAND)

        self.__contentSizer.Add(self.__btnSizer, flag=wx.EXPAND)

        self.__mainSizer.Add(self.__icon,
                             flag=wx.ALL | wx.CENTRE,
                             border=20)
        self.__mainSizer.Add(self.__contentSizer,
                             flag=wx.EXPAND | wx.ALL,
                             proportion=1,
                             border=20)

        self.__message.Wrap(self.GetSize().GetWidth())

        yes  = self.__yesButton
        no   = self.__noButton
        cncl = self.__cancelButton

        if   focus == 'yes':                         yes .SetDefault()
        elif focus == 'no'     and no   is not None: no  .SetDefault()
        elif focus == 'cancel' and cncl is not None: cncl.SetDefault()

        self.SetSizer(self.__mainSizer)
        self.Layout()
        self.Fit()
        self.CentreOnParent()


    def CheckBoxState(self, index=0):
        """After this ``CheckBoxMessageDialog`` has been closed, this method
        will retrieve the state of the dialog ``CheckBox``.
        """
        return self.__checkboxes[index].GetValue()


    def __onYesButton(self, ev):
        """Called when the button on this ``CheckBoxMessageDialog`` is
        clicked. Closes the dialog.
        """
        self.EndModal(wx.ID_YES)


    def __onNoButton(self, ev):
        """Called when the button on this ``CheckBoxMessageDialog`` is
        clicked. Closes the dialog.
        """
        self.EndModal(wx.ID_NO)


    def __onCancelButton(self, ev):
        """If the ``CHECKBOX_MSGDLG_CANCEL_BUTTON`` style was set, this method
        is called when the cancel button is clicked. Closes the dialog.
        """
        self.EndModal(wx.ID_CANCEL)



#!/usr/bin/env python
#
# colourbarbitmap.py - A function which renders a colour bar using
# matplotlib as an RGBA bitmap.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides a single function, :func:`colourBarBitmap`, which uses
:mod:`matplotlib` to plot a colour bar. The colour bar is rendered off-screen
and returned as an RGBA bitmap.
"""


def colourBarBitmap(cmap,
                    width,
                    height,
                    cmapResolution=256,
                    negCmap=None,
                    invert=False,
                    gamma=1,
                    ticks=None,
                    ticklabels=None,
                    tickalign=None,
                    label=None,
                    orientation='vertical',
                    labelside='top',
                    alpha=1.0,
                    fontsize=10,
                    bgColour=None,
                    textColour='#ffffff',
                    scale=1.0):
    """Plots a colour bar using :mod:`matplotlib`.


    The rendered colour bar is returned as a RGBA bitmap within a
    ``numpy.uint8`` array of size :math:`w \\times h \\times 4`, with the
    top-left pixel located at index ``[0, 0, :]``.


    A rendered colour bar will look something like this:

    .. image:: images/colourbarbitmap.png
       :scale: 50%
       :align: center


    :arg cmap:           Name of a registered :mod:`matplotlib` colour map.

    :arg width:          Colour bar width in pixels.

    :arg height:         Colour bar height in pixels.

    :arg cmapResolution: Colour map resolution (number of distinct colours).

    :arg negCmap:        If provided, two colour maps are drawn, centered at 0.

    :arg invert:         If ``True``, the colour map is inverted.

    :arg gamma:          Gamma correction factor - exponentially weights the
                         colour map scale towards one end.

    :arg ticks:          Locations of tick labels. Ignored if
                         ``ticklabels is None``.

    :arg ticklabels:     Tick labels.

    :arg tickalign:      Tick alignment (one for each tick, either ``'left'``,
                         ``'right'``, or ``'center'``).

    :arg label:          Text label to show next to the colour bar.

    :arg orientation:    Either ``vertical`` or ``horizontal``.

    :arg labelside:      Side of the colour bar to put the label - ``top``,
                         ``bottom``, ``left`` or ``right``. If
                         ``orientation='vertical'``, then ``top``/``bottom``
                         are interpreted as ``left``/``right`` (and vice-versa
                         when ``orientation='horizontal'``).

    :arg alpha:          Colour bar transparency, in the range ``[0.0 - 1.0]``.

    :arg fontsize:       Label font size in points.

    :arg bgColour:       Background colour - can be any colour specification
                         that is accepted by :mod:`matplotlib`.

    :arg textColour:     Label colour - can be any colour specification that
                         is accepted by :mod:`matplotlib`.

    :arg scale:          DPI scaling factor.
    """

    # These imports are expensive, so we're
    # importing at the function level.
    import numpy                           as np
    import matplotlib.backends.backend_agg as mplagg
    import matplotlib.figure               as mplfig

    if orientation not in ['vertical', 'horizontal']:
        raise ValueError('orientation must be vertical or '
                         'horizontal ({})'.format(orientation))

    if orientation == 'horizontal':
        if   labelside == 'left':  labelside = 'top'
        elif labelside == 'right': labelside = 'bottom'
    else:
        if   labelside == 'top':    labelside = 'left'
        elif labelside == 'bottom': labelside = 'right'

    if labelside not in ['top', 'bottom', 'left', 'right']:
        raise ValueError('labelside must be top, bottom, left '
                         'or right ({})'.format(labelside))

    # vertical plots are rendered horizontally,
    # and then simply rotated at the end
    if orientation == 'vertical':
        width, height = height, width
        if labelside == 'left': labelside = 'top'
        else:                   labelside = 'bottom'

    # Default is 96 dpi to an inch
    winches   = width  / 96.0
    hinches   = height / 96.0
    dpi       = scale  * 96.0
    ncols     = cmapResolution
    data      = genColours(cmap, ncols, invert, alpha, gamma)

    if negCmap is not None:
        ndata  = genColours(negCmap, ncols, not invert, alpha, gamma)
        data   = np.concatenate((ndata, data), axis=1)
        ncols *= 2

    # force tick positions to
    # the left edge of the
    # corresponding colour
    if ticks is not None:
        ticks = [t - 0.5 / ncols for t in ticks]

    fig     = mplfig.Figure(figsize=(winches, hinches), dpi=dpi)
    canvas  = mplagg.FigureCanvasAgg(fig)
    ax      = fig.add_subplot(111)

    if bgColour is not None:
        fig.patch.set_facecolor(bgColour)
    else:
        fig.patch.set_alpha(0)

    # draw the colour bar
    ax.imshow(data,
              aspect='auto',
              origin='lower',
              interpolation='nearest')

    ax.set_yticks([])
    ax.tick_params(colors=textColour, labelsize=fontsize, length=0)

    if labelside == 'top':
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position('top')
    else:
        ax.xaxis.tick_bottom()
        ax.xaxis.set_label_position('bottom')

    if label is not None:
        ax.set_xlabel(label, fontsize=fontsize, color=textColour)
        label = ax.xaxis.get_label()

    if ticks is None or ticklabels is None:
        ax.set_xticks([])
    else:

        ax.set_xticks(np.array(ticks) * ncols)
        ax.set_xticklabels(ticklabels)
        ticklabels = ax.xaxis.get_ticklabels()

    # Resize the colour bar to make
    # space for the ticks and labels
    wpad        = 5 / float(width)
    hpad        = 5 / float(height)
    totalHeight = 1.0

    # Figure out text height in
    # pixels (with 6 pixel padding)
    fontpix    = (fontsize / 72.0) * dpi / scale
    textHeight = (fontpix + 6) / float(height)

    if label      is not None: totalHeight -= textHeight
    if ticklabels is not None: totalHeight -= textHeight

    if labelside == 'top': bottom, top = 0, totalHeight
    else:                  bottom, top = 1 - totalHeight, 1

    fig.subplots_adjust(
        left=wpad,
        right=1.0 - wpad,
        bottom=bottom + hpad,
        top=top - hpad)

    # Divide the vertical height between the
    # colour bar, the label, and the tick labels.
    if label is not None:

        # I don't understand why, but I have
        # to set va to the opposite of what
        # I would have thought.
        if labelside == 'top':
            label.set_va('bottom')
            label.set_position((0.5, 1.0))
        else:
            label.set_va('top')
            label.set_position((0.5, 0))

    # Adjust tick label horizontal alignment. This
    # must be done *after* calling tick_top/tick_bottom,
    # as I think the label objects get recreated.
    if ticklabels is not None and tickalign is not None:
        for l, a in zip(ticklabels, tickalign):
            l.set_horizontalalignment(a)

    ax.set_xlim((-0.5, ncols - 0.5))

    canvas.draw()

    buf = canvas.tostring_argb()
    ncols, nrows = canvas.get_width_height()

    bitmap = np.frombuffer(buf, dtype=np.uint8)
    bitmap = bitmap.reshape(nrows, ncols, 4).transpose([1, 0, 2])

    # the bitmap is in argb order,
    # but we want it in rgba
    rgb = bitmap[:, :, 1:]
    a   = bitmap[:, :, 0]
    bitmap = np.dstack((rgb, a))

    if orientation == 'vertical':
        bitmap = np.flipud(bitmap.transpose([1, 0, 2]))
        bitmap = np.rot90(bitmap, 2)

    return bitmap


def genColours(cmap, cmapResolution, invert, alpha, gamma=1):
    """Generate an array containing ``cmapResolution`` colours from the given
    colour map object/function.
    """

    import numpy         as np
    import matplotlib.cm as cm

    ncols = cmapResolution
    cmap  = cm.get_cmap(cmap)
    idxs  = np.linspace(0.0, 1.0, ncols)

    if gamma != 1:
        idxs = idxs ** gamma

    if invert:
        idxs = idxs[::-1]

    cmap       = cmap(idxs)
    cmap[:, 3] = alpha

    return np.dstack((cmap, cmap)).transpose((2, 0, 1))



#!/usr/bin/env python
#
# widgetlist.py - A widget which displays a list of groupable widgets.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WidgetList` class, which displays a list
of widgets.
"""


from collections import OrderedDict

import wx
import wx.lib.newevent      as wxevent
import wx.lib.scrolledpanel as scrolledpanel

from . import togglepanel


class WidgetList(scrolledpanel.ScrolledPanel):
    """A scrollable list of widgets.

    The ``WidgetList`` provides a number of features:

      - Widgets can be grouped.
      - A label can be shown next to each widget.
      - Widget groups can be collapsed/expanded.
      - Widgets and groups can be dynamically added/removed.


    The most important methods are:

     .. autosummary::
        :nosignatures:

        AddWidget
        AddGroup


    A ``WidgetList`` looks something like this:

     .. image:: images/widgetlist.png
        :scale: 50%
        :align: center


    A ``WidgetList`` emits a :data:`WidgetListChangeEvent` whenever its
    contents change.
    """


    _defaultOddColour   = '#eaeaea'
    """Background colour for widgets on odd rows. """


    _defaultEvenColour  = '#ffffff'
    """Background colour for widgets on even rows. """


    _defaultGroupColour = '#cdcdff'
    """Border and title background colour for widget groups. """


    def __init__(self, parent, style=0, minHeight=-1):
        """Create a ``WidgetList``.

        :arg parent:    The :mod:`wx` parent object.
        :arg style:     Passed through to ``wx.ScrolledPanel.__init__``
        :arg minHeight: Minimum height of each row
        """
        self.__minHeight   = minHeight
        self.__widgSizer   = wx.BoxSizer(wx.VERTICAL)
        self.__sizer       = wx.BoxSizer(wx.VERTICAL)

        self.__groupSizer  = wx.BoxSizer(wx.VERTICAL)
        self.__widgets     = OrderedDict()
        self.__groups      = OrderedDict()

        self.__oddColour   = WidgetList._defaultOddColour
        self.__evenColour  = WidgetList._defaultEvenColour
        self.__groupColour = WidgetList._defaultGroupColour

        self.__sizer.Add(self.__widgSizer,  flag=wx.EXPAND)
        self.__sizer.Add(self.__groupSizer, flag=wx.EXPAND)

        self.__oneExpanded = style & WL_ONE_EXPANDED

        # The SP.__init__ method seemingly
        # induces a call to DoGetBestSize,
        # which assumes that all of the
        # things above exist. So we call
        # init after we've created those
        # things.
        scrolledpanel.ScrolledPanel.__init__(self, parent)

        self.SetSizer(self.__sizer)
        self.SetBackgroundColour((255, 255, 255))
        self.SetupScrolling()
        self.SetAutoLayout(1)


    def DoGetBestSize(self):
        """Returns the best size for the widget list, with all group
        widgets expanded.
        """

        width, height = self.__widgSizer.GetSize().Get()
        for name, group in self.__groups.items():
            w, h    = group.parentPanel.GetBestSize().Get()
            w      += 20
            h      += 10

            if w > width:
                width = w
            height += h

        return wx.Size(width, height)


    def __makeWidgetKey(self, widget):
        """Widgets are stored in a dictionary - this method generates a
        string to use as a key, based on the widget ``id``.
        """
        return str(id(widget))


    def __setLabelWidths(self, widgets):
        """Calculates the maximum width of all widget labels, and sets all
        labels to that width.

        This ensures that all labels/widgets line are horizontally aligned.
        """

        if len(widgets) == 0:
            return

        dc        = wx.ClientDC(widgets[0].label)
        lblWidths = [dc.GetTextExtent(w.displayName)[0] for w in widgets]
        maxWidth  = max(lblWidths)

        for w in widgets:
            w.label.SetMinSize((maxWidth + 10, -1))
            w.label.SetMaxSize((maxWidth + 10, -1))


    def __setColours(self):
        """Called whenever the widget list needs to be refreshed.

        Makes sure that odd/even widgets and their labels have the correct
        background colour.
        """
        def setWidgetColours(widgDict):
            for i, widg in enumerate(widgDict.values()):

                if i % 2: colour = self.__oddColour
                else:     colour = self.__evenColour
                widg.SetBackgroundColour(colour)

        setWidgetColours(self.__widgets)

        for group in self.__groups.values():

            setWidgetColours(group.widgets)
            group.parentPanel.SetBackgroundColour(self.__groupColour)
            group.colPanel   .SetBackgroundColour(self.__groupColour)


    def __refresh(self, *args, **kwargs):
        """Updates widget colours (see :meth:`__setColours`), and lays out
        the widget list.

        :arg postEvent: If ``True`` (the default), a
                        :data:`WidgetListChangeEvent` is posted.
        """
        self.__setColours()
        self.FitInside()
        self.Layout()

        if kwargs.get('postEvent', True):
            wx.PostEvent(self, WidgetListChangeEvent())


    def SetColours(self, odd=None, even=None, group=None):
        """Sets the colours used on this ``WidgetList``.

        Each argument is assumed to be a tuple of ``(r, g, b)`` values,
        each in the range ``[0 - 255]``.

        :arg odd:   Background colour for widgets on odd rows.
        :arg even:  Background colour for widgets on even rows.
        :arg group: Border/title colour for widget groups.
        """
        if odd   is not None: self.__oddColour   = odd
        if even  is not None: self.__evenColour  = even
        if group is not None: self.__groupColour = group
        self.__setColours()


    def GetGroups(self):
        """Returns a list containing the name of every group in this
        ``WidgetList``.
        """
        return list(self.__groups.keys())


    def HasGroup(self, groupName):
        """Returns ``True`` if this ``WidgetList`` contains a group
        with the specified name.
        """
        return groupName in self.__groups


    def RenameGroup(self, groupName, newDisplayName):
        """Changes the display name of the specified group.

         .. note:: This method only changes the *display name* of a group,
                   not the group identifier name. See the :meth:`AddGroup`
                   method.

        :arg groupName:      Name of the group.

        :arg newDisplayName: New display name for the group.
        """
        group = self.__groups[groupName]
        group.displayName = newDisplayName
        group.colPanel.SetLabel(newDisplayName)


    def AddGroup(self, groupName, displayName=None):
        """Add a new group to this ``WidgetList``.

        A :exc:`ValueError` is raised if a group with the specified name
        already exists.

        :arg groupName:   The name of the group - this is used as an
                          identifier for the group.

        :arg displayName: A string to be shown in the title bar for the
                          group. This can be changed later via the
                          :meth:`RenameGroup` method.
        """

        if displayName is None:
            displayName = groupName

        if groupName in self.__groups:
            raise ValueError('A group with name {} '
                             'already exists'.format(groupName))

        parentPanel = wx.Panel(self, style=wx.SUNKEN_BORDER)
        colPanel    = togglepanel.TogglePanel(parentPanel, label=displayName)

        widgPanel   = colPanel.GetPane()
        widgSizer   = wx.BoxSizer(wx.VERTICAL)

        widgPanel.SetSizer(widgSizer)

        gapSizer = wx.BoxSizer(wx.VERTICAL)

        # A spacer exists at the top,
        # and between, every group.
        gapSizer.Add((-1, 5))
        gapSizer.Add(parentPanel, border=10, flag=(wx.EXPAND |
                                                   wx.LEFT   |
                                                   wx.RIGHT))
        self.__groupSizer.Add(gapSizer, flag=wx.EXPAND)

        parentSizer = wx.BoxSizer(wx.VERTICAL)
        parentSizer.Add(colPanel,
                        border=5,
                        flag=wx.EXPAND | wx.BOTTOM,
                        proportion=0)
        parentPanel.SetSizer(parentSizer)

        group = _Group(groupName,
                       displayName,
                       gapSizer,
                       parentPanel,
                       colPanel,
                       widgPanel,
                       widgSizer)

        self.__groups[groupName] = group
        self.__refresh()

        # Mouse wheel listener needed
        # on all children under linux/GTK
        if wx.Platform == '__WXGTK__':
            parentPanel.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)
            colPanel   .Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        colPanel.Bind(togglepanel.EVT_TOGGLEPANEL_EVENT, self.__onGroupExpand)


    def GetWidgets(self, groupName=None):
        """Returns a list containing all of the widgets that have been added
        to this ``WidgetList``.

        :arg groupName: If provided, only widgets in the specified group will
                        be returned. Otherwise, ungrouped widgets are returned.
        """
        if groupName is None: widgDict = self.__widgets
        else:                 widgDict = self.__groups[groupName].widgets
        widgets = [w.widget for w in widgDict.values()]
        return widgets


    def AddWidget(self, widget, displayName, tooltip=None, groupName=None):
        """Add an arbitrary widget to the property list.

        If the ``groupName`` is not provided, the widget is added to a list
        of *top level* widgets, which appear at the top of the list, above
        any groups. Otherwise, the widget is added to the collapsible panel
        corresponding to the specified group.

        A :exc:`ValueError` is raised if the widget is already contained
        in the list.

        :arg widget:      The widget to add to the list.

        :arg displayName: The widget label/display name.

        :arg tooltip:     A tooltip for the widget.

        :arg groupName:   Name of the group to which the widget should be
                          added.

         .. note:: The provided ``widget`` may also be a :class:`wx.Sizer`
                   instances, although support for this is basic. Specifically,
                   only one level of nesting is possible, i.e. the provided
                   ``wx.Sizer`` may not have any other ``wx.Sizer``
                   instances as its children.
        """

        if groupName is None:
            widgDict    = self.__widgets
            parent      = self
            parentSizer = self.__widgSizer
        else:
            group       = self.__groups[groupName]
            widgDict    = group.widgets
            parent      = group.widgPanel
            parentSizer = group.sizer

        key = self.__makeWidgetKey(widget)

        if key in widgDict:
            raise ValueError('Widgets {} already exist'.format(key))

        widgPanel = wx.Panel(parent)
        widgSizer = wx.BoxSizer(wx.HORIZONTAL)
        widgPanel.SetSizer(widgSizer)

        if isinstance(widget, wx.Sizer):
            for child in widget.GetChildren():
                child.GetWindow().Reparent(widgPanel)
        else:
            w, h = widget.GetBestSize().Get()
            if self.__minHeight > h:
                h = self.__minHeight
            widget.SetMinSize( (w, h))
            widget.Reparent(widgPanel)

        label = wx.StaticText(widgPanel,
                              label=displayName,
                              style=wx.ALIGN_RIGHT)

        widgSizer.Add(label,  flag=wx.EXPAND)
        widgSizer.Add(widget, flag=wx.EXPAND, proportion=1)

        parentSizer.Add(widgPanel,
                        flag=wx.EXPAND | wx.LEFT | wx.RIGHT,
                        border=5)

        widg = _Widget(displayName,
                       tooltip,
                       label,
                       widget,
                       widgPanel,
                       widgSizer)

        if tooltip is not None:
            widg.SetTooltip(tooltip)

        # Under linux/GTK, mouse events are
        # captured by child windows, so if
        # we want scrolling to work, we need
        # to capture scroll events on every
        # child. Under OSX/cocoa, this is
        # not necessary.
        if wx.Platform == '__WXGTK__':
            widg.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        widgDict[key] = widg

        self.__setLabelWidths(list(widgDict.values()))
        self.__refresh()


    def __onMouseWheel(self, ev):
        """Only called if running on GTK. Scrolls the widget list according
        to the mouse wheel rotation.
        """

        posx, posy = self.GetViewStart()
        rotation   = ev.GetWheelRotation()

        if   rotation > 0: delta =  5
        elif rotation < 0: delta = -5
        else:              return

        if ev.GetWheelAxis() == wx.MOUSE_WHEEL_VERTICAL: posy -= delta
        else:                                            posx += delta

        self.Scroll(posx, posy)


    def __onGroupExpand(self, ev):
        """Called when the user expands or collapses a group. Enforces
        the :data:`WL_ONE_EXPANDED` style if it is enabled, and refreshes
        the panel.
        """

        panel = ev.GetEventObject()

        if panel.IsExpanded() and self.__oneExpanded:
            for group in self.__groups.values():
                if group.colPanel is not panel:
                    group.colPanel.Collapse()

        self.__refresh()


    def AddSpace(self, groupName=None):
        """Adds some empty vertical space to the widget list.

        :arg groupName: Name of the group tio which the space should be added.
                        If not specified, the space is added to the *top level*
                        widget list - see the :meth:`AddWidget` method.
        """
        if groupName is None: parentSizer = self.__widgSizer
        else:                 parentSizer = self.__groups[groupName].sizer

        parentSizer.Add((-1, 10))


    def RemoveWidget(self, widget, groupName=None):
        """Removes and destroys the specified widget from this ``WidgetList``.

        :arg widget:    The widget to remove.

        :arg groupName: Name of the group in which the widget is contained.
        """
        key = self.__makeWidgetKey(widget)

        if groupName is None:
            parentSizer = self.__widgSizer
            widgDict    = self.__widgets
        else:
            group       = self.__groups[groupName]
            parentSizer = group.sizer
            widgDict    = group.widgets

        widg = widgDict.pop(key)
        parentSizer.Detach(widg.panel)

        widg.Destroy()
        self.__refresh()


    def RemoveGroup(self, groupName):
        """Removes the specified group, and destroys all of the widgets
        contained within it.
        """
        group = self.__groups.pop(groupName)

        self.__groupSizer.Detach(group.gapSizer)
        group.parentPanel.Destroy()
        self.__refresh()


    def Clear(self):
        """Removes and destroys all widgets and groups. """
        for key in list(self.__widgets.keys()):
            widg = self.__widgets.pop(key)
            self.__widgSizer.Detach(widg.sizer)
            widg.Destroy()

        for group in self.GetGroups():
            self.RemoveGroup(group)
        self.__refresh()


    def ClearGroup(self, groupName):
        """Removes and destroys all widgets in the specified group, but
        does not remove the group.
        """
        group = self.__groups[groupName]
        group.sizer.Clear(True)
        group.widgets.clear()
        self.__refresh()


    def GroupSize(self, groupName):
        """Returns the number of widgets that have been added to the
        specified group.
        """
        return len(self.__groups[groupName].widgets)


    def IsExpanded(self, groupName):
        """Returns ``True`` if the panel for the specified group is currently
        expanded, ``False`` if it is collapsed
        """
        return self.__groups[groupName].colPanel.IsExpanded()


    def Expand(self, groupName, expand=True):
        """Expands or collapses the panel for the specified group. """
        panel = self.__groups[groupName].colPanel
        if expand: panel.Expand()
        else:      panel.Collapse()
        self.__refresh()


class _Widget(object):
    """The ``_Widget`` class is used internally by the :class:`WidgetList`
    to organise references to each widget in the list.
    """
    def __init__(self,
                 displayName,
                 tooltip,
                 label,
                 widget,
                 panel,
                 sizer):
        self.displayName = displayName
        self.tooltip     = tooltip
        self.label       = label
        self.widget      = widget
        self.panel       = panel
        self.sizer       = sizer


    def SetBackgroundColour(self, colour):
        self.panel.SetBackgroundColour(colour)
        self.label.SetBackgroundColour(colour)

        if isinstance(self.widget, wx.Sizer):
            for c in self.widget.GetChildren():
                c.GetWindow().SetBackgroundColour(colour)
        else:
            self.widget.SetBackgroundColour(colour)


    def SetTooltip(self, tooltip):

        self.label.SetToolTip(wx.ToolTip(tooltip))

        if isinstance(self.widget, wx.Sizer):
            for child in self.widget.GetChildren():
                child.GetWindow().SetToolTip(wx.ToolTip(tooltip))
        else:
            self.widget.SetToolTip(wx.ToolTip(tooltip))


    def Bind(self, evType, callback):
        self.panel.Bind(evType, callback)
        self.label.Bind(evType, callback)

        if isinstance(self.widget, wx.Sizer):
            for c in self.widget.GetChildren():
                c.GetWindow().Bind(evType, callback)
        else:
            self.widget.Bind(evType, callback)


    def Destroy(self):
        self.label.Destroy()
        if isinstance(self.widget, wx.Sizer):
            self.widget.Clear(True)
        else:
            self.widget.Destroy()


class _Group(object):
    """The ``_Group`` class is used internally by :class:`WidgetList`
    instances to represent groups of widgets that are in the list.
    """
    def __init__(self,
                 groupName,
                 displayName,
                 gapSizer,
                 parentPanel,
                 colPanel,
                 widgPanel,
                 sizer):
        self.groupName   = groupName
        self.displayName = displayName
        self.gapSizer    = gapSizer
        self.parentPanel = parentPanel
        self.colPanel    = colPanel
        self.widgPanel   = widgPanel
        self.sizer       = sizer
        self.widgets     = OrderedDict()



_WidgetListChangeEvent, _EVT_WL_CHANGE_EVENT = wxevent.NewEvent()


WidgetListChangeEvent = _WidgetListChangeEvent
"""Event emitted by a :class:`WidgetList` when its contents change. """


EVT_WL_CHANGE_EVENT = _EVT_WL_CHANGE_EVENT
"""Identifier for the :data:`WidgetListChangeEvent`. """


WL_ONE_EXPANDED = 1
""":class:`WidgetList` style flag. When applied, at most one group will
be expanded at any one time.
"""



#!/usr/bin/env python
#
# __init__.py - Small utility modules for doing random things.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package contains a collection of small utility modules for doing random
things. A few functions are also defined at the package level:

.. autosummary::
   :nosignatures:

    wxversion
    isalive
"""



WX_PYTHON  = 1
"""Constant returned by the ``wxversion`` function, indicating that wxPython
3.0.2.0 or older is being used.
"""


WX_PHOENIX = 2
"""Constant returned by the ``wxversion`` function, indicating that
wxPython/Phoenix (>= 3.0.3) is being used.
"""


def wxversion():
    """Determines the version of wxPython being used. Returns either ``WX_PYTHON``
    or ``WX_PHOENIX``.
    """

    import wx

    pi        = [t.lower() for t in wx.PlatformInfo]
    isPhoenix = False

    for tag in pi:
        if 'phoenix' in tag:
            isPhoenix = True
            break

    if isPhoenix: return WX_PHOENIX
    else:         return WX_PYTHON


def isalive(widget):
    """Returns ``True`` if the given ``wx.Window`` object is "alive" (i.e.  has
    not been destroyed), ``False`` otherwise. Works in both wxPython and
    wxPython/Phoenix.

    .. warning:: Don't try to test whether a ``wx.MenuItem`` has been
                 destroyed, as it will probably result in segmentation
                 faults. Check the parent ``wx.Menu`` instead.
    """

    import wx

    wxver = wxversion()

    if wxver == WX_PHOENIX:
        excTypes = (RuntimeError,)
    else:
        excTypes = (wx.PyDeadObjectError,)

    try:
        if hasattr(widget, 'IsBeingDeleted'):
            return not widget.IsBeingDeleted()
        else:
            widget.GetParent()
            return True

    except excTypes:
        return False



#!/usr/bin/env python
#
# __init__.py - fsleyes-widgets
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``fsleyes_widgets`` package contains various custom :mod:`wx` widgets
and utilities used by FSLeyes.


Some of the controls in ``fsleyes_widgets`` are duplicates of controls which
are already available in ``wx`` or ``wx.lib.agw``. In these instances, I wrote
my own implementations to work around annoying, quirky, and/or downright buggy
behaviour in the existing controls.


This file is used to store the current ``fsleyes-widgets`` version.
"""


__version__ = '0.8.3'


from fsleyes_widgets.utils import (WX_PYTHON,  # noqa
                                   WX_PHOENIX,
                                   wxversion,
                                   isalive)



#!/usr/bin/env python
#
# test_elistbox.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import random

import wx

from . import run_with_wx, simclick, simtext, simkey, realYield

from fsleyes_widgets import isalive
import fsleyes_widgets.elistbox as elistbox


def test_create():
    run_with_wx(_test_create)
def _test_create():
    frame = wx.GetApp().GetTopWindow()

    items    = random.sample(range(100), 5)
    labels   = [str(i)                 for i in items]
    tooltips = ['tip: {}'.format(i)    for i in items]
    listbox  = elistbox.EditableListBox(
        frame,
        labels=labels,
        clientData=items,
        tooltips=tooltips)

    assert listbox.VisibleItemCount() == 5
    assert listbox.GetCount()         == 5

    assert listbox.GetSelection() == wx.NOT_FOUND
    listbox.SetSelection(3)
    assert listbox.GetSelection() == 3

    for idx, i in enumerate(items):
        assert listbox.IndexOf(i) == idx

    assert list(listbox.GetLabels()) == list(labels)
    assert list(listbox.GetData())   == list(items)
    for i in range(len(items)):
        assert listbox.GetItemTooltip(i) == tooltips[i]
        assert listbox.GetItemLabel(  i) == labels[i]
        assert listbox.GetItemData(   i) == items[i]

def test_add_remove():
    run_with_wx(_test_add_remove)
def _test_add_remove():
    frame   = wx.GetApp().GetTopWindow()
    items   = [str(random.randint(1, 100)) for i in range(5)]
    listbox = elistbox.EditableListBox(frame, items)

    # delete
    for i in range(5):
        ridx = random.randint(0, len(items) - 1)
        items.pop(ridx)
        listbox.Delete(ridx)
        assert list(listbox.GetLabels()) == items

    # append
    for i in range(5):
        item = str(random.randint(0, 100))
        items.append(item)
        listbox.Append(item)
        assert list(listbox.GetLabels()) == items

    # insert
    for i in range(5):
        item = str(random.randint(0, 100))
        idx  = random.randint(0, len(items) - 1)
        items.insert(idx, item)
        listbox.Insert(item, idx)
        assert list(listbox.GetLabels()) == items
